// Copyright (c) 2000 by the University of Southern California
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.


#include <template.h>
#include <caf/caf.h>
#include <caf/cntable.h>
#include <random.h>
#include <address.h>
#include <mobilenode.h>
#include <god.h>
#include <phy.h>
#include <wireless-phy.h>
#include <energy-model.h>



int hdr_caf::offset_;
static class CAFHeaderClass : public PacketHeaderClass {
public:
	CAFHeaderClass() : PacketHeaderClass("PacketHeader/CAF",
					     sizeof(hdr_caf)) {
		bind_offset(&hdr_caf::offset_);
	}
} class_cafhdr;

static class CAFAgentClass : public TclClass {
public:
	CAFAgentClass() : TclClass("Agent/CAF") {}
	TclObject* create(int argc, const char*const* argv) {
	        assert(argc == 5);
		return (new CAFAgent((nsaddr_t) atoi(argv[4])));
	}
} class_cafagent;

static class CAFPartnerClass : public TclClass {
public:
        CAFPartnerClass() : TclClass("CAFPartner") {}
        TclObject* create(int, const char*const*) {
                return (new CAFPartner());
        }
} class_cafparnter;


static inline double 
cafjitter (double max, int be_random_)
{
  return (be_random_ ? Random::uniform(max) : max);
}

static inline double
cafjitter2 () {
  return(Random::uniform(0,DISCOVER_SPACE));
}



CAFAgent::CAFAgent(nsaddr_t id) : Agent(PT_CAF), beacon_(1), randomflag_(1),leader_starttime_(0),timer_(this), stimer_(this), dtimer_(this), sptimer_(this),grtimer_(this),state_(CAF_FREE)
{

	nid_ = id;
	cid_ = -1;
	nlen_ = 0;
	mywaketime_ = 0;
	gstate_ = CAF_NON_GATEWAY;
	adjclusters_ = 0;
	caf_nosleep_ = 0;
	thisnode = Node::get_node_by_address(nid_); 
	init_eng_ = 0;
	threshold_eng_ = 0;

	table_ = new CAFnTable;

}

void CAFAgent::recv(Packet* p, Handler *)
{
	hdr_caf *cafh = hdr_caf::access(p);

	switch (cafh->type_) {

	case CAF_DISCOVER:

	  if (state_ != CAF_SLEEP) {

	      processDiscoveryMsg(p);
	       
	  }

	  //printf("node %d at state %d at time %f\n",nid_, state_, Scheduler::instance().clock());
	  Packet::free(p);
	  break;

	case CAF_GATEWAY:
	  if (state_ == CAF_LEADER && cid_ != nid_ && gstate_ == CAF_NON_PRIMARY_GATEWAY) {
	      processGatewayMsg(p);
	  }
	  Packet::free(p);
	  break;

	default:
	  Packet::free(p);
	  break;
	}
}

void CAFAgent::processGatewayMsg(Packet* p)
{
  u_int32_t  i,j,removeflag, prigtyflag;
  u_int32_t  gnid;
  int32_t    gcid;
  u_int32_t  gnlen; 
  u_int32_t  gnhead[NMAX];

	u_int32_t dst;

	unsigned char *w = p->accessdata ();

	// neighbor'd nid
	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
	
	gnid = dst;

	// neighbor's cid
	
	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);

	gcid = dst;

	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);

	gnlen = dst;

	for (i=0; i<gnlen; i++) {
	    
	  dst = *(w++);
	  dst = dst << 8 | *(w++);
	  dst = dst << 8 | *(w++);
	  dst = dst << 8 | *(w++);
	  gnhead[i] = dst;

	}


	//printf("Node (%d %d) got GATEWAY msg from (%d %d)\n", nid_, cid_, gnid, gcid);

	// decide what to do

	// check to see if my cid is in the cluster head's neighbor
	if (cid_ != gcid) {
	  prigtyflag = 0;
	  for (i=0; i<gnlen; i++) {
	    if (cid_ == (int32_t)gnhead[i]) {
	      
	      prigtyflag = 1;
	      break;
	    }
	    
	  }
	}


	// if mycid == this primary gateway node cid

	if (cid_ == gcid || prigtyflag) {
	  // find if my adjacent list is contained by the primary
	  // list

	  for (i=0; i<gnlen; i++) {
	    removeflag = 0;
	    for (j=0; j<adjclusters_; j++) {

	      if (adjclusterlist_[j] == (int32_t)gnhead[i]) {
		removeflag = 1;
		break;
	      }

	    }

	    if (1 == removeflag) {
	      for (;j<adjclusters_; j++) {
		adjclusterlist_[j] = adjclusterlist_[j+1];
	      }
	      adjclusters_--;
	    }
	    
	  }

	  removeflag = 0;
	  if (prigtyflag) {
	    if (adjclusters_ == 1) {
	      if ((int32_t)gnhead[0] == gcid) {
		removeflag = 1;
	      }

	    }
	  }
	  
	  if (adjclusters_ == 0 || removeflag) {
	        // go to sleep now
		dtimer_.force_cancel();
		mywaketime_ = leader_starttime_ - NOW;
		if (mywaketime_ < 0) mywaketime_ = 0;
		
		//printf("Node %d gave up NON_PRI_GATEWAY role at %f\n", nid_, Scheduler::instance().clock());

		schedule_wakeup();

	  }

	 
	}
}


void CAFAgent::processDiscoveryMsg(Packet* p)
{
  double ttl;
  //	struct DiscoveryMsg emsg;
        cntable_ent ent;
	u_int32_t i;
	u_int32_t dst;

	unsigned char *w = p->accessdata ();
	
	// neighbor'd nid
	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
	
	ent.nid = dst;

	// neighbor's cid
	
	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);

	ent.cid = dst;

	// neighbor's state

	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
	
	ent.state = dst;

	// neighbor's enlt

	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);

	ent.enlt = dst;

	// neighbor's stime

	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
	
	ent.stime = dst;

	// neighbor's neighbor len
	
	dst = *(w++);
	dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);
        dst = dst << 8 | *(w++);

	ent.nlen = dst;

	// fill the node's neighbor list 
	for (i=0; i<ent.nlen; i++) {
	  
	  dst = *(w++);
	  dst = dst << 8 | *(w++);
	  dst = dst << 8 | *(w++);
	  dst = dst << 8 | *(w++);
	  ent.nlist[i] = dst;

	}
	
	table_->AddEntry(ent);

	// if I am a leader but catch a discovery msg from another
	// leader, I should change my state depending on who is the
	// the winner

	if (state_ == CAF_LEADER && cid_ == (int32_t)nid_ ) {
	    if (ent.state == CAF_LEADER && ent.cid == (int32_t)ent.nid) {
	        ttl = leader_starttime_-NOW;
		if (ttl < 0) ttl = 0;

		if (ttl > ent.enlt) {
		    send_discovery();
		    return;
		} else {
		    
		    if (ttl == ent.enlt && nid_ < ent.nid) {
		        send_discovery();
			return;
		    }
		}
		// go to sleep now
		dtimer_.force_cancel();
		sptimer_.force_cancel();
		leader_starttime_ = 0;
		mywaketime_ = ent.enlt;
		cid_ = ent.cid;
		//printf("Node %d gave up LEADER role at %f\n", nid_, Scheduler::instance().clock());

		schedule_wakeup();
	    } else {
	        send_discovery();
	    }

	}


}

void CAFAgent::formCluster() {
  cntable_ent * pcte = NULL;
  u_int32_t i;
  int clusterneighbor;
  /*
   * If I am clustered, check my neighbor to
   * see if anybody has better lifetime than me
   * If YES, change my clusterid
   */

  // mywaketime_ = enlt_;

  // If I have decided myself as cluster head, keep it.
  if (cid_ == (int32_t)nid_) return;
 
  // If I am already assigned a cluster head, look to see
  // if I can find a better one

  if (cid_ > -1) {
      pcte = table_->FirstEntry();
      while (pcte) {
	if (pcte->nid == (u_int32_t) pcte->cid) {
	    if (pcte->enlt > mywaketime_ 
		|| (pcte->enlt == mywaketime_ && cid_ > pcte->cid)) {
	      
	            cid_ = pcte->cid;
	            mywaketime_= pcte->enlt;
	    }

	    // build my neighbor head list

	    for (i=0; i<nlen_; i++) {
	      if (nhead_[i] == (u_int32_t)pcte->cid) break;
	    }

	    if (i == nlen_ && nlen_ < 20) {
	        nhead_[i] = pcte->cid;
	        nlen_++;
	    }
	}

	pcte = table_->NextEntry();
	
      }
     
      return;
  }

  /* 
   * Now I am not clustered. Decide if I can become a head
   */
  
  pcte = table_->FirstEntry();
  
  clusterneighbor = 0;


  while (pcte) {

      // build my neighbor head list

      if (pcte->cid == (int32_t)pcte->nid) {
	  
	  for (i=0; i<nlen_; i++) {
	      if (nhead_[i] == (u_int32_t)pcte->cid) break;
	  }

	  if (i == nlen_ && nlen_ < 20) {
	        nhead_[i] = pcte->cid;
	        nlen_++;
	  }

	  clusterneighbor = 1;

      }

      pcte = table_->NextEntry();
  }

  /* 
   * I am not clustered, but I have cluster head neighbor
   */
  
  //mywaketime_ = 0;

  if (clusterneighbor) {
      
    for (i = 0; i < nlen_; i++) {
      pcte = table_->GetEntry(nhead_[i]);
      if (pcte->enlt > mywaketime_ || (pcte->enlt == mywaketime_ && pcte->cid < cid_)) {
	cid_ = pcte->cid;
	mywaketime_ = pcte->enlt;
      }
    }
    return;


  } else {

    /* if I do not have the cluster head neighbor, check if I have
     * the max lifetime in all of my unclustered neighbors.
     */
    pcte = table_->FirstEntry();

    while (pcte) {

      // my neighbor is a cluster head, I fail
      if ((u_int32_t)pcte->cid == pcte->nid) break;

      // my neighbor is unclustered but has longer network liftime, I fail
      if (pcte->cid < 0 && pcte->enlt > enlt_) break;
      
      // my neighbor is unclustered, have the same emlt with me,
      // my lower node id, I fail
      
      if (pcte->cid < 0 && pcte->enlt == enlt_ && pcte->nid < nid_) break;

      pcte = table_->NextEntry();
    }
    
    if (!pcte) {
      cid_ = nid_;
      return;
    }

  }

}

bool CAFAgent::selectGateway() 
{
    cntable_ent * pcte = NULL;
    u_int32_t  i,j;
    int firstmatch, secondmatch;

    if ((u_int32_t)cid_ == nid_) 
        return FALSE; // I am the cluster leader, no way to be gateway


    // if I do not have > 1 cluster header neighbors, I am not gateway nodes
    // FIXME: this only means I am NOT the primary gateway nodes.
    // what about non-gateway nodes.

    if (nlen_ == 1) {
  
      // search my neighbor list to see if they have
      // neighbors from other cluster
      
      pcte = table_->FirstEntry();

      while (pcte) {

	  if (pcte->cid != cid_) {
	      gstate_ = CAF_NON_PRIMARY_GATEWAY;
	      return TRUE;
	  }
	  pcte = table_->NextEntry();
      }
      
      return FALSE;

    }

    // check my leader neighbor list first, then check my non-leader neighbors
    // to see if my leader neighbors are subset of them, if yes, I am not
    // the gateway winner. Otherwise, I am the gateway winner

    pcte = table_->FirstEntry();

    while (pcte) {
      secondmatch = 0;        
      for (i = 0; i< nlen_; i++) {

	  firstmatch = 0;
	  for (j=0; j<pcte->nlen; j++) {
	      if (nhead_[i] == pcte->nlist[j]) {
		  firstmatch = 1;
		  break;
	      }
	  }

	  if (firstmatch == 0) break;

	  secondmatch ++;

      }
    
      if ((u_int32_t)secondmatch == nlen_) {

	if (nlen_ < pcte->nlen) return FALSE;  // my head neighbor is a subset
	if (nlen_ == pcte->nlen) {
	  // compare who has longer enlt
	  if (enlt_ < pcte->enlt) return FALSE;
	  if (enlt_ == pcte->enlt && nid_ > pcte->nid) return FALSE;

	}

      }

      pcte = table_->NextEntry();

    }
    
    gstate_ = CAF_PRIMARY_GATEWAY;
    return TRUE;

}


void CAFAgent::schedule_wakeup() {

  cntable_ent * cent;
  node_off();

  // decide how long I have to sleep
  // get my cluster header first

  cent = table_->GetEntry(cid_);

  if (!cent) {
    printf("Node %d can not find its cid_ %d\n", nid_, cid_);
    exit(1);
  }

  mywaketime_ = cent->enlt;

  if (cent->stime < cent->enlt) { 

    //printf("Node %d's cluster head stime=%d, enlt=%d\n", nid_, cent->stime, cent->enlt );
      mywaketime_ = cent->stime;

  }

  //printf("Node %d goes sleep for %f at %f\n", nid_, mywaketime_, Scheduler::instance().clock());

  dtimer_.resched(mywaketime_);

}

double CAFAgent::myenlt()
{
  double ce,maxp;
  Phy *phyp; 
  double enlt; 

  ce = (thisnode->energy_model())->energy();

  if (caf_nosleep_) {

    ce = threshold_eng_ - (init_eng_-ce);
    if (ce < 0) ce = 0;

  }

  phyp = (thisnode->ifhead()).lh_first;

  if (phyp) {
	      maxp = ((WirelessPhy *)phyp)->getPtconsume();
  } 

  enlt = ce/maxp;
  
  return enlt;

}

// all the non-primary gateway node to build 
// up its adjacent tables.
void
CAFAgent::gateway_build_adjacent() 
{

      u_int32_t i, insertflag;
      cntable_ent * pcte = NULL;

      pcte = table_->FirstEntry();

      while (pcte) {

	  if (pcte->cid != cid_ && pcte->cid != -1 && cid_ != -1) {
	    insertflag = 0;
	    for (i=0; i<adjclusters_; i++) {
	      if (adjclusterlist_[i] == pcte->cid) {
		insertflag = 1;
		break;
	      }
	    }
	    if (0 == insertflag) {
	      adjclusterlist_[adjclusters_] = pcte->cid;
	      adjclusters_++;
	    }
	  }
	    
	  pcte = table_->NextEntry();
      }


      // debug
      //printf ("Node (%d %d) has ADJ list:", nid_, cid_);
      //for (i=0; i<adjclusters_; i++) {
      //printf("%d ",adjclusterlist_[i]);
      //}

      //printf("\n");
      


      

}

// timeout process for discovery phase
// the same timeout call back for different timeout events
void CAFAgent::timeout(CafMsgType msgt)
{
  double nexttime;
  
    switch (msgt) {
    case CAF_MSG_DISCOVER: // Discover timeout events
        switch (state_) {
	// node is in sleep state, ignore the discover timeout events
	case CAF_FREE:      // node is in discovery state
	    if (discover_starttime_ < DISCOVER_MAXTIMES) {
	        send_discovery();
		discover_starttime_++;
		nexttime = discover_starttime_*DISCOVER_SPACE+discover_starttime_offset_-NOW;
		nexttime = Random::uniform(0, nexttime);
		
		//printf("node %d resched DISCOVER timer at %f for %f\n", nid_, NOW, nexttime); 
	      
		timer_.resched(nexttime);
	    } 
	    break;
	default:
	    break;
	}
	break; // end of processing of discover timeout events

    case CAF_MSG_SELECT: // cluster select timer events
        switch (state_) {
	case CAF_FREE:
	    setCAFstate(CAF_CLUSTER);
	    cluster_starttime_ = 0;
	    cluster_starttime_offset_ = NOW;
	    stimer_.resched(cafjitter2());
	    
	    dtimer_.resched(DUTY_STARTTIME);
	    break;

	case CAF_CLUSTER:
	    if (cluster_starttime_ < (2*DISCOVER_MAXTIMES) ) {
	        formCluster();
		send_discovery();
		cluster_starttime_++;
		
		nexttime = cluster_starttime_*DISCOVER_SPACE+cluster_starttime_offset_-NOW;
		nexttime = Random::uniform(0, nexttime);
		//printf("node %d resched DISCOVER(CLUSTER) timer at %f for %f\n", nid_, NOW, nexttime); 

		stimer_.resched(nexttime);
	    } 
	    break;
	default:
	    break;
	}
	break; // end of processing of caf selecting events

    case CAF_MSG_DUTY: // leader timesout
        switch (state_) {
	case CAF_CLUSTER:
	    formCluster();
	    if (-1 == cid_) {
	        cid_ = nid_;      // I am still not clustered on time
		
		// I need to start a special timer to
		// tell other people about myself.
		//printf("Node %d starts sp timer at %f\n", nid_, Scheduler::instance().clock());		
		sptimer_.resched(Random::uniform(0,1));
	    }

	  // if I am the cluster head
	    if ((u_int32_t)cid_ == nid_) {
	        setCAFstate(CAF_LEADER);
		leader_starttime_ = NOW + enlt_;
		//table_-> cleanTable();

		dtimer_.resched(enlt_);

		//printf("Node %d becomes head at time %f\n", nid_, Scheduler::instance().clock());
		
	    } else {
	      
	        if (selectGateway()) {
		  //int i;
		    setCAFstate(CAF_LEADER);
		    leader_starttime_ = NOW + mywaketime_;

		    //printf("Node (%d %d) becomes gateway of ", nid_, cid_);

		    //for (i=0; i<(int)nlen_; i++) {
		    //printf("%d ", nhead_[i]);

		    //}
		    // printf("\n");

		    // build up gateway adjacent table
		    if (gstate_ == CAF_NON_PRIMARY_GATEWAY) {
		        gateway_build_adjacent();
		    }
		    //table_-> cleanTable();
		    dtimer_.resched(mywaketime_);

		    // start gateway re-selection process
		    if (gstate_ == CAF_PRIMARY_GATEWAY) {
		        grtimer_.resched(Random::uniform(0,3));
		    }

		} else {

		  //printf("Node %d goes sleep for cluster %d\n", nid_, cid_); 

		    schedule_wakeup();
		}

	    }

	    break;
	case CAF_LEADER:
	case CAF_SLEEP:
	
	    //setCAFstate(CAF_FREE);
	    sptimer_.force_cancel();
	    table_-> cleanTable(); 
	    cid_ = -1; // uncluster
	    nlen_ = 0;
	    mywaketime_ = 0;
	    
	    // set enlt_
	    if ((enlt_ = (u_int32_t)myenlt()) > MIN_LIFETIME) {
	        enlt_ = (u_int32_t) enlt_/2;
	    }

	    if (enlt_ > 0 || caf_nosleep_) {
	        node_on();
		gstate_ = CAF_NON_GATEWAY;
		adjclusters_ = 0;
		// start over again
		timer_.resched(cafjitter2());
		discover_starttime_ = 0;
		discover_starttime_offset_ = NOW;

		// start select timer in 1 second from now
		stimer_.resched(DISCOVER_STARTTIME);
		cluster_starttime_ = 0;
	    } else {
	        node_off();
	    }
	    
	    break;
	default:
	    break;

	}
	break; // end of processing of caf leader timer events
    case CAF_MSG_SPDUTY: 
        if (state_ == CAF_LEADER) {
            send_discovery();
            sptimer_.resched(Random::uniform(0,1));
        } 
        break;
    case CAF_MSG_GATEWAY:
        if (state_ == CAF_LEADER && gstate_ == CAF_PRIMARY_GATEWAY) {
	    send_gatewayinfo();
	    grtimer_.resched(Random::uniform(0,3));
	  
        }
	break;
    default:
        break;
    }

}
	  
int CAFAgent::command(int argc, const char*const* argv)
{
        if (argc == 2) {
	  if (strcmp (argv[1], "start-caf") == 0) {
	    // randomly schedule 5 discovery msg within 
	    // 1 second, then timeouts for state change

	    //printf("node %d resched DISCOVER timer at %f\n", nid_, NOW);

	    timer_.resched(cafjitter2());
	    discover_starttime_ = 0;
	    discover_starttime_offset_ = NOW;

	    // start select timer in 1 second from now

	    stimer_.resched(DISCOVER_STARTTIME);
	    cluster_starttime_ = 0;

	    init_eng_ = (thisnode->energy_model())->energy();

	    // set enlt_
	    if ((enlt_ = (u_int32_t)myenlt()) > MIN_LIFETIME) {
	        enlt_ = (u_int32_t) enlt_/2;
	    }

	    return (TCL_OK); 
	  }
	  if (strcmp (argv[1], "no-sleep") == 0) {
	    caf_nosleep_ = 1;
	    return (TCL_OK); 
	  }
	}

	if (argc == 3) {
	    if (strcmp (argv[1], "threshold") == 0) {
	      threshold_eng_ = atof(argv[2]);
	      return (TCL_OK); 
	    }
	}

	return (Agent::command(argc, argv));
}

void CAFAgent::send_gatewayinfo()
{
        Packet *p = allocpkt();

	hdr_caf *h = hdr_caf::access(p);
	//hdr_ip *iph = hdr_ip::access(p);

	h->type_ = CAF_GATEWAY;
	
	makeUpGatewayMsg(p);

	send(p,0);  

}

// primary gateway -> non-primary-gateway
void
CAFAgent::makeUpGatewayMsg(Packet *p)
{
  int i;
  hdr_ip *iph = hdr_ip::access(p);
  hdr_cmn *hdrc = hdr_cmn::access(p);
  unsigned char *walk;

  // fill up the header
  hdrc->next_hop_ = IP_BROADCAST;
  hdrc->addr_type_ = NS_AF_INET;
  iph->daddr() = IP_BROADCAST << Address::instance().nodeshift();
  iph->dport() = 254;

  p->allocdata(sizeof(GatewayMsgHead));
    // fill up the data

  walk = p->accessdata ();
  hdrc->size_ = sizeof(GatewayMsgHead) + IP_HDR_LEN; // Existence Msg + IP

  // fill the gatewaymsg head
  
  // this node
  *(walk++) = nid_ >> 24;
  *(walk++) = (nid_ >> 16) & 0xFF;
  *(walk++) = (nid_ >> 8) & 0xFF;
  *(walk++) = (nid_ >> 0) & 0xFF;

  // my cluster id
  *(walk++) = cid_ >> 24;
  *(walk++) = (cid_ >> 16) & 0xFF;
  *(walk++) = (cid_ >> 8) & 0xFF;
  *(walk++) = (cid_ >> 0) & 0xFF;

  // my cluster neighbor len
  *(walk++) = nlen_ >> 24;
  *(walk++) = (nlen_ >> 16) & 0xFF;
  *(walk++) = (nlen_ >> 8) & 0xFF;
  *(walk++) = (nlen_ >> 0) & 0xFF;

  for (i=0; i<(int)nlen_; i++) {
    *(walk++) = nhead_[i] >> 24;
    *(walk++) = (nhead_[i] >> 16) & 0xFF;
    *(walk++) = (nhead_[i] >> 8) & 0xFF;
    *(walk++) = (nhead_[i] >> 0) & 0xFF;
  }

}


void CAFAgent::send_discovery()
{
        Packet *p = allocpkt();

	hdr_caf *h = hdr_caf::access(p);
	//hdr_ip *iph = hdr_ip::access(p);

	h->type_ = CAF_DISCOVER;
	// h->seqno_ = ++seqno_;

	// update my grid infomation
	
	makeUpDiscoveryMsg(p);

	send(p,0);
}

void
CAFAgent::makeUpDiscoveryMsg(Packet *p)
{
  hdr_ip *iph = hdr_ip::access(p);
  hdr_cmn *hdrc = hdr_cmn::access(p);
  u_int32_t i,stime, enlt2;
  unsigned char *walk;
  double gridsize, speed;


  // fill up the header
  hdrc->next_hop_ = IP_BROADCAST;
  hdrc->addr_type_ = NS_AF_INET;
  iph->daddr() = IP_BROADCAST << Address::instance().nodeshift();
  iph->dport() = 254;

  p->allocdata(sizeof(DiscoveryMsgHead));
    // fill up the data

  walk = p->accessdata ();
  hdrc->size_ = sizeof(DiscoveryMsgHead) + IP_HDR_LEN; // Existence Msg + IP

  // fill the discoverymsg head

  // this node
  *(walk++) = nid_ >> 24;
  *(walk++) = (nid_ >> 16) & 0xFF;
  *(walk++) = (nid_ >> 8) & 0xFF;
  *(walk++) = (nid_ >> 0) & 0xFF;

  // my cluster id
  *(walk++) = cid_ >> 24;
  *(walk++) = (cid_ >> 16) & 0xFF;
  *(walk++) = (cid_ >> 8) & 0xFF;
  *(walk++) = (cid_ >> 0) & 0xFF;

  // my state
  *(walk++) = state_ >> 24;
  *(walk++) = (state_ >> 16) & 0xFF;
  *(walk++) = (state_ >> 8) & 0xFF;
  *(walk++) = (state_ >> 0) & 0xFF;  

  // enlt tells the receiver that how much longer the sender can
  // survive, cut it into half for the purpose of load balance

  if (state_ == CAF_LEADER) {
      // must send real msg because I am the leader
	enlt_ = (u_int32_t) (leader_starttime_ - NOW);
	if (enlt_ < 0) enlt_ = 0;

  } 
  //else {
  //
  //  if ((enlt = (u_int32_t)myenlt()) > MIN_LIFETIME) {
  //      enlt = (u_int32_t) enlt/2;
  //  }
  //}

  enlt2 = (u_int32_t) enlt_;

  *(walk++) = enlt2 >> 24;
  *(walk++) = (enlt2 >> 16) & 0xFF;
  *(walk++) = (enlt2 >> 8) & 0xFF;
  *(walk++) = (enlt2 >> 0) & 0xFF;

  // fill my possible time of leaving this grid
  
  speed = ((MobileNode*)thisnode)->speed();

  if (speed == 0) {
    // make stime big enough
    stime = 2*enlt2;
  } else {
    gridsize = 60;  // 250m is the transmission radius, reduce it if not
                     // the result is not good.
    stime = (u_int32_t) (gridsize/speed);
  }

  *(walk++) = stime >> 24;
  *(walk++) = (stime >> 16) & 0xFF;
  *(walk++) = (stime >> 8) & 0xFF;
  *(walk++) = (stime >> 0) & 0xFF;

  // how many cluster head neighbor do I have

  *(walk++) = nlen_ >> 24;
  *(walk++) = (nlen_ >> 16) & 0xFF;
  *(walk++) = (nlen_ >> 8) & 0xFF;
  *(walk++) = (nlen_ >> 0) & 0xFF;

  // if there are some neighbors, fill them in

  for (i=0; i<nlen_; i++) {
    *(walk++) = nhead_[i] >> 24;
    *(walk++) = (nhead_[i] >> 16) & 0xFF;
    *(walk++) = (nhead_[i] >> 8) & 0xFF;
    *(walk++) = (nhead_[i] >> 0) & 0xFF;
  }
}


void
CAFAgent::node_off()
{
    Phy *p;
    EnergyModel *em;

    if (caf_nosleep_) {
        setCAFstate(CAF_LEADER);
        return;
    }

    //printf ("Node (%d %d) goes SLEEP from %d at %f\n",nid_, cid_, state_, Scheduler::instance().clock());

    // set node state
    em = thisnode->energy_model();
    em->node_on() = false;

    // notify phy
    p  = (thisnode->ifhead()).lh_first;
    if (p) {
	((WirelessPhy *)p)->node_off();
    }
    // change agent state
    setCAFstate(CAF_SLEEP);
}

void
CAFAgent::node_on()
{
    Phy *p;
    EnergyModel* em;

    // set node state

    em = thisnode->energy_model();
    em->node_on() = true;

    //(MobileNode *)thisnode->energy_model()->node_on() = true;

    // notify phy

    p = (thisnode->ifhead()).lh_first;
    if (p) {
        ((WirelessPhy *)p)->node_on();
    }
    
    setCAFstate(CAF_FREE);
}

void 
CAFAgent::setCAFstate(CafNodeState gs)
{
  //printf("Node (%d %d) changes state from %d to %d at %f\n", nid_, cid_, state_,gs,Scheduler::instance().clock());

  printf("%f %d %d %d\n",Scheduler::instance().clock(), nid_, state_, gs);   

   state_ = gs;
}

CAFPartner::CAFPartner() : Connector(), cafagent_(1),mask_(0xffffffff),
        shift_(8)
{
        bind("addr_", (int*)&(here_.addr_));
        bind("port_", (int*)&(here_.port_));
        bind("shift_", &shift_);
        bind("mask_", &mask_);
}

void CAFPartner::recv(Packet* p, Handler *h)
{
        hdr_ip* hdr = hdr_ip::access(p);
	hdr_cmn *hdrc = hdr_cmn::access(p);

	/* handle PT_CAF packet only */
	/* convert the dst ip if it is -1, change it into node's */
	/* own ip address */	
	if ( hdrc->ptype() == PT_CAF ) {
	  if (cafagent_ == 1) {
	    if (hdr->daddr() == IP_BROADCAST) {
		hdr->daddr() = here_.addr_;
	    }	    
	  } else {
	    /* if cafagent is not installed, drop the packet */
	    drop (p);
	    return;
	  }
	}
	
	target_->recv(p,h);


}

int CAFPartner::command(int argc, const char*const* argv)
{
        if (argc == 3) {
	  if (strcmp (argv[1], "set-cafagent") == 0) {
	    cafagent_ = atoi(argv[2]);
	    return (TCL_OK); 
	  }
	}
	return Connector::command(argc, argv); 
}

void CAFDiscoverTimer::expire(Event *) {
  a_->timeout(CAF_MSG_DISCOVER);
}

void CAFSelectTimer::expire(Event *) {
  a_->timeout(CAF_MSG_SELECT);
}

void CAFDutyTimer::expire(Event *) {
  a_->timeout(CAF_MSG_DUTY);
}

void CAFSPDutyTimer::expire(Event *) {
  a_->timeout(CAF_MSG_SPDUTY);
}

void CAFGRTimer::expire(Event *) {
  a_->timeout(CAF_MSG_GATEWAY);
}
