
#include <stdlib.h>
 
#include "random.h"
#include "trafgen.h"
#include "ranvar.h"

#define INTERVAL802 0.00011
#define INTERVAL1290 0.0002

/* 
 * Mpeg traffic source based on switched ethernet traffic of SunRay 
 * Pameterized by interval, (optional) random noise in the interval  
 */

class MPEG_Traffic : public TrafficGenerator {
 public:
	MPEG_Traffic();
	virtual double next_interval(int&);
	//HACK so that udp agent knows interpacket arrival time within a burst
	inline double interval() { return (interval_); }
 protected:
	virtual void start();
	void init();
	double rate_;     /* send rate during on time (bps) */
	double interval_; /* packet inter-arrival time during burst (sec) */
	double random_;
	int seqno_;
	int maxpkts_;
	int w802_;
	int w1290_;
	int maxlen1290_;
	int maxlen802_;
	int cur_run_;
	int w0_cnt_;
	
};


static class MPEGTrafficClass : public TclClass {
 public:
	MPEGTrafficClass() : TclClass("Application/Traffic/MPEG") {}
	TclObject* create(int, const char*const*) {
		return (new MPEG_Traffic());
	}
} class_mpeg_traffic;

MPEG_Traffic::MPEG_Traffic() : seqno_(0)
{
	//bind_bw("rate_", &rate_);
	//bind("random_", &random_);
	//bind("packetSize_", &size_);
	//bind("maxpkts_", &maxpkts_);
}

void MPEG_Traffic::init()
{

	// initial size = 802 and interval = 0.00011
	// how many 802-lengh packets

	size_ = 802;
	interval_ = (double)0.00011;
	maxlen802_ = (int)Random::uniform(1, 100);
	cur_run_ = 802;

	maxlen1290_ = (int)Random::uniform(1, 10);

	// define weight

	w802_ = 90;
	w1290_ = 9;

	w0_cnt_ = 0;

	if (agent_)
		agent_->set_pkttype(PT_CBR);
}

void MPEG_Traffic::start()
{
        init();
        running_ = 1;
        timeout();
}

double MPEG_Traffic::next_interval(int& size)
{

	switch (cur_run_) {
	  case 802:

	    size_= 802;
	    interval_ = INTERVAL802;
	    
	    w802_--;
	    maxlen802_--;

	    // decide the max pkts of next time 802 sequence.
	    // Use WRR so that if w802 is positive, more credits
	    // otherwise, less credits.

	    if (maxlen802_ == 0) {
		
		// try to balance the weight, allow 10% wave

		w802_ = 90 + w802_;
		maxlen802_ = (int)(w802_ * (1 + Random::uniform(-0.1,0.1)));
		
		// pass to 1290 packets
		cur_run_ = 1290;
		
		// change packetsize and interval
		size_ = 1290;
		interval_ = Random::uniform(0.001,0.01);
		
	    }
	    
	    break;

	  case 1290:

	    size_ = 1290;
	    interval_ = INTERVAL1290;
	    
	    w1290_--;
	    maxlen1290_--;

	    if (maxlen1290_ == 0) {
		
		// try to balance the weight, allow 10% wave

		w1290_ = 9 + w1290_;
		maxlen1290_ = (int)(w1290_ * (1 + Random::uniform(-0.5,0.5)));
	
		w0_cnt_++;

		if (w0_cnt_ > 20) {
		
		    // pass to random packets
		    cur_run_ = 0;
		    w0_cnt_ = 0;

		    // change packetsize and interval
		    size_ = (int)(Random::uniform(50,1500));
		    interval_ = Random::uniform(0.001,0.01);
		} else {
	
		    cur_run_ = 802;
	            // change packetsize and interval
                    size_ = 802;
                    interval_ = Random::uniform(0.0001,0.1);                    

		}
	    }
	    break;
	    
	  case 0:

	    // simply switch to 820 packets
	    cur_run_ = 802;
	    size_ = 802;
	    interval_ = Random::uniform(0.0001,0.1);
	    break;
	    
	  default:

	    printf("Error in mpeg traffic generator\n");
	    break;
	}
	
	// Recompute interval in case rate_ or size_ has changes

	// interval_ = (double)(size_ << 3)/(double)rate_;

	double t = interval_;

	// if (random_)
	// 	t += interval_ * Random::uniform(-0.5, 0.5);	
	// size = size_;
	//if (++seqno_ < maxpkts_)
	//	return(t);
	//else
	//	return(-1);
	size = size_;
	return(t);
}


